from os.path import realpath

from cronspell.cronspell import Cronspell


def locate():
    """
    * finds the location of the meta model. For your convenience
    """
    print(realpath(Cronspell().meta_model_src))  # noqa: T201



"""Ultimate Notion provides a pythonic, high-level API for Notion

Notion-API: https://developers.notion.com/reference/intro
"""

from importlib.metadata import PackageNotFoundError, version

from cronspell.resolve import resolve as cronspell

try:
    __version__ = version("cronspell")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


__all__ = ["cronspell"]



import typer
import typer.rich_utils

from cronspell import __version__
from cronspell.cli import locate, parse, preflight, to_dot

app = typer.Typer(
    name=f"CronSpell {__version__}",
    help="Date-expression domain specific language parsing. "
    'A neat way to express things like "First Saturday of any year", '
    'or "3rd thursdays each month" and such',
    pretty_exceptions_enable=False,
)
app.command(name="parse")(parse.parse)
app.command(name="dot")(to_dot.to_dot)
app.command(name="locate")(locate.locate)
app.command(name="preflight")(preflight.preflight)

if __name__ == "__main__":
    app()



import enum
import logging
import sys
from pathlib import Path

from yamlpath import Processor
from yamlpath.common import Parsers


class LogLevel(str, enum.Enum):
    CRITICAL = "critical"
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"
    DEBUG = "debug"


def setup_logging(log_level: LogLevel):
    """Setup basic logging"""
    log_format = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    numeric_level = getattr(logging, log_level.upper(), None)
    logging.basicConfig(
        level=numeric_level,
        stream=sys.stdout,
        format=log_format,
        datefmt="%Y-%m-%d %H:%M:%S",
    )


yaml = Parsers.get_yaml_editor()


def get_processor(file: Path) -> Processor:
    logging.raiseExceptions = True
    logging.logThreads = False

    log = logging.getLogger("root")
    yaml_file = file
    (yaml_data, _) = Parsers.get_yaml_data(yaml, log, yaml_file)
    return Processor(log, yaml_data)



import hashlib
from pathlib import Path
from typing import Annotated

import typer
from textx.export import model_export

from cronspell.cronspell import Cronspell


def to_dot(
    expressions: Annotated[
        list[str],
        typer.Argument(
            ...,
            help="One or more Date-Expressions, e.g 'now /month'",
        ),
    ],
    out: Annotated[
        Path,
        typer.Option("--out", "-o", show_default=False, help="Where to write output"),
    ],
    sha_len: Annotated[
        int,
        typer.Option(
            "--sha-len",
            "-s",
            show_default=True,
            help="Number of character to truncate the sha to",
            default_factory=lambda: 7,
        ),
    ],
    pad_len: Annotated[
        int,
        typer.Option(
            "--pad-len",
            "-p",
            show_default=True,
            help="Number of character to pad the leading number to",
            default_factory=lambda: 3,
        ),
    ],
):
    """
    * From a list of valid expressions, generate graphviz dot diagrams.
    * Writes to `--out`.
    * Filenames are based on sha and position in list.
    """

    cronspell = Cronspell()
    pad = max(pad_len, len(str(len(expressions))))

    for idx, expression in enumerate(expressions):
        model = cronspell.meta_model.model_from_str(expression)
        sha = hashlib.sha3_224(f"{expression.replace(' ', '')}".encode()).hexdigest()[0:sha_len]
        destination = Path.joinpath(out, f"{idx + 1:0{pad}}{'_' if sha_len > 0 else ''}{sha}.dot")
        model_export(model, destination)



from pathlib import Path
from typing import Annotated

import typer
from yamlpath.exceptions import UnmatchedYAMLPathException

from cronspell.cli.yaml import get_processor
from cronspell.resolve import resolve


class CronspellPreflightException(BaseException): ...


def preflight(
    files: Annotated[
        list[Path],
        typer.Argument(
            ...,
            help="One or more Paths.",
        ),
    ],
    yamlpath: Annotated[
        str,
        typer.Option("--yamlpath", "-p", show_default=False, help="yamlpath YAML_PATH"),
    ],
):
    """
    * Takes a list of paths
    * validates expressions
    """

    for file in files:
        processor = get_processor(file)

        try:
            for token in processor.get_nodes(yamlpath, mustexist=True):
                resolve(str(token).strip())
        except UnmatchedYAMLPathException as ex:
            msg = f"yamlpath {yamlpath} does not exist in {file}!"
            raise CronspellPreflightException(file, msg) from ex



# SPDX-FileCopyrightText: 2024-present iilei • jochen preusche <922226+iilei@users.noreply.github.com>
#
# SPDX-License-Identifier: MIT


import calendar
import functools
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional
from zoneinfo import ZoneInfo

from textx import metamodel_from_file

from cronspell.exceptions import CronpellMathException

with calendar.different_locale(locale=("EN_US", "UTF-8")):
    # first day of the week: Monday as per ISO Standard
    WEEKDAYS = [*calendar.day_abbr]


TIME_UNITS_SHORT = ["Y", "m", "CW", "W", "d", "H", "M", "S"]
TIME_RESETS = [
    ("year", 1970),
    ("month", 1),
    None,
    None,
    ("day", 1),
    ("minute", 0),
    ("hour", 0),
    ("second", 0),
]
TIME_RESETS_MAP = [[TIME_UNITS_SHORT[k], v] for k, v in enumerate(TIME_RESETS) if v]
FLOOR_CW_MAX = 52
FLOOR_Y_MAX = 9999


class Cronspell:
    def __init__(self, timezone: Optional[ZoneInfo] = None):
        if timezone is None:
            timezone = ZoneInfo("UTC")

        self.meta_model_src = Path.joinpath(Path(__file__).parent, "cronspell.tx")
        self.meta_model = metamodel_from_file(self.meta_model_src, use_regexp_group=True)
        self.timezone = timezone

    def parse_anchor(self):
        if (anchor := getattr(self.model, "anchor", None)) and ((tznow := anchor.tznow) or (isodate := anchor.isodate)):
            return datetime.now(ZoneInfo(tznow.tz)) if (tznow and tznow.tz) else datetime.fromisoformat(isodate)

        return datetime.now(self.timezone)

    @staticmethod
    def get_time_unit(alias):
        return next(name for name in [*TIME_UNITS_SHORT, *WEEKDAYS] if getattr(alias, name, None))

    def step(self, current, step):  # noqa: PLR0914
        # operation ~> Minus|Plus|Floor|Ceil
        operation = step.statement._tx_fqn.rpartition(".")[-1]

        # Year Modulo
        if operation == "YModulo":
            resolution = max(step.statement.value, 1)

            if resolution > FLOOR_Y_MAX:
                msg = f"Year Modulo needed lower than {FLOOR_Y_MAX + 1}! Got {resolution}."
                raise CronpellMathException(msg)

            year = current.isocalendar().year // resolution * resolution

            return datetime(year, 1, 1, tzinfo=self.timezone)

        # Calendar Week Modulo
        elif operation == "CwModulo":
            resolution = max(step.statement.value, 1)

            if resolution > FLOOR_CW_MAX:
                msg = f"Calendar Week Modulo needed lower than {FLOOR_CW_MAX + 1}! Got {resolution}."
                raise CronpellMathException(msg)

            week = current.isocalendar().week // resolution * resolution
            y, m, d, *_ = datetime.strptime(f"{current.isocalendar().year}-W{week}" + "-1", "%Y-W%W-%w").timetuple()

            return datetime(y, m, d, tzinfo=self.timezone)
        else:
            resolution = step.statement.res._tx_fqn.rpartition(".")[-1]
            time_unit = self.get_time_unit(step.statement.res)

        if operation in {"Plus", "Minus"}:
            # special case: week ~> 7 days
            if time_unit == "W":
                delta = (datetime.strptime("2", "%d") - datetime.strptime("1", "%d")) * (
                    (-7 if operation == "Minus" else 7) * step.statement.steps
                )
            else:
                lower_value = [x for x in TIME_RESETS_MAP if x[0] == time_unit].pop()[1][1]

                delta = (
                    datetime.strptime(str(lower_value + 2), f"%{time_unit}")
                    - datetime.strptime(str(lower_value + 1), f"%{time_unit}")
                ) * ((-1 if operation == "Minus" else 1) * step.statement.steps)

            return current + delta

        elif operation == "Floor" and resolution == "WeekDay":
            offset_abs = (7 + (current.weekday() - WEEKDAYS.index(time_unit))) % 7
            offset = -1 * offset_abs if operation == "Floor" else 7 - offset_abs
            current += timedelta(days=offset)

            # operation "Floor" to be performed as per day
            time_unit = "d"

        if operation == "Floor":
            prune = TIME_UNITS_SHORT[TIME_UNITS_SHORT.index(time_unit) + 1 :]

            current = current.replace(**dict([x[1] for x in TIME_RESETS_MAP if x[0] in prune]))

            return current

    def parse(self, expression: str = "now") -> datetime:
        self.expression = expression
        self.model = self.meta_model.model_from_str(expression)
        self.anchor = self.parse_anchor()

        self.tz = self.anchor.tzname()

        result = functools.reduce(self.step, [*getattr(self.model, "date_math_term", [])], self.anchor)

        return result



from typing import Annotated

import typer

from cronspell import cronspell


def parse(
    expression: Annotated[
        str,
        typer.Argument(
            help="Date-Expression, e.g 'now /month",
            default_factory=lambda: "now",
        ),
    ],
    fmt: Annotated[
        str,
        typer.Option(
            "--format", "-f", show_default=False, help="Optional format parameter.", default_factory=lambda: ""
        ),
    ],
):
    """
    * Turn a valid expression into a datetime.
    * ISOFormat or as per `--format` argument.
        * Format templating: https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior
    """
    parsed = cronspell(expression)
    print(parsed.strftime(fmt) if len(fmt) > 0 else parsed.isoformat())  # noqa: T201
